from cobo_custody.config.env import Env

SANDBOX_ENV = Env(host="https://api.sandbox.cobo.com",
                  coboPub="032f45930f652d72e0c90f71869dfe9af7d713b1f67dc2f7cb51f9572778b9c876")
DEVELOP_ENV = Env(host="https://api.develop.cobo.com",
               coboPub="03596da539963fb1dd29d5859e25903eb76b9f7ed2d58516e29c9f80c201ff2c1b")
PROD_ENV = Env(host="https://api.custody.cobo.com",
               coboPub="02c3e5bacf436fbf4da78597e791579f022a2e85073ae36c54a361ff97f2811376")

SANDBOX_TEST_DATA = {
    "cobo_id": "20220314181458000331767000003732",
    "tx_id": "0x1c4d137bc2a2ee8f22cbdf9e90405974e72e65d922f42eb81d9f7a05d0f64fc6",
    "withdraw_id": "web_send_by_user_915_1647252768642",
    "deposit_address": {"BTC": "3JBYNrbB4bHtGWHTEa3ZPuRK9kwTiEUo4D",
                        "XRP": "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|2047482901"},
    "deposit_addresses": {"BTC": "3JBYNrbB4bHtGWHTEa3ZPuRK9kwTiEUo4D,bc1qf22hpu33u2tkyy528mdvpnre45n8lu5s3ycatu",
                          "XRP": "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|2047482901,"
                                 "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|3752417374"},
    "loop_address": {"BTC": "35eXJPLRTSp4Wn8n2f6pkQF4t3KdU2cuhz",
                     "XRP": "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|477817505"},
    "loop_addresses": {"BTC": "35eXJPLRTSp4Wn8n2f6pkQF4t3KdU2cuhz,34R4JHecUwGNEFVGKz1vR8R6BHGi5FUqPt",
                       "XRP": "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|477817505,"
                              "rfKyCMyoV6Ln2GZ7YDbrBrnXCbAyBbxRqB|2874421071"}
    }

PROD_TEST_DATA = {
    "cobo_id": "20220311154108000184408000002833",
    "tx_id": "4041A888C9966BE8916FE65F2FEE7AE9A9DC3F49D0F1643A768C842CA95FA736",
    "withdraw_id": "sdk_request_id_fe80cc5f_1647068483396",
    "deposit_address": {"BTC": "36xYx7vf7DUKpJDixpY3EoV2jchFwYSNCb",
                        "XRP": "rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|3992922539"},
    "deposit_addresses": {"BTC": "36xYx7vf7DUKpJDixpY3EoV2jchFwYSNCb,bc1q0l24tf5sjdu9t7l6hrlhxz9aq9yeej9h2sc7tk",
                          "XRP": "rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|3992922539,"
                                 "rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|1492154866"},
    "loop_address": {"BTC": "34WLjtk9ta96BVxc1jRF7j5eVvehoftsVV",
                     "XRP": "rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|633829231"},
    "loop_addresses": {"BTC": "34WLjtk9ta96BVxc1jRF7j5eVvehoftsVV,33P1kjMfDCKipR58S7XbsCqbmPT5YGrhUo",
                       "XRP": "rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|633829231,rBWpYJhuJWBPAkzJ4kYQqHShSkkF3rgeD|935940214"}
    }



import json
import time
from typing import Tuple
from urllib.parse import urlencode

import requests

from cobo_custody.client.api_response import ApiResponse
from cobo_custody.config import Env
from cobo_custody.error.api_error import ApiError
from cobo_custody.signer.api_signer import ApiSigner
from cobo_custody.signer.local_signer import verify_ecdsa_signature


class MPCClient(object):
    def __init__(self, signer: ApiSigner, env: Env, debug: bool = False):
        self.api_signer = signer
        self.env = env
        self.debug = debug

    def sort_params(self, params: dict) -> str:
        params = [(key, val) for key, val in params.items()]

        params.sort(key=lambda x: x[0])
        return urlencode(params)

    def remove_none_value_elements(self, input_dict: dict) -> dict:
        if type(input_dict) is not dict:
            return {}
        result = {}
        for key in input_dict:
            tmp = {}
            if input_dict[key] is not None:
                if type(input_dict[key]).__name__ == 'dict':
                    tmp.update({key: self.remove_none_value_elements(input_dict[key])})
                else:
                    tmp.update({key: input_dict[key]})
            result.update(tmp)
        return result

    def verify_response(self, response: requests.Response) -> Tuple[bool, dict]:
        content = response.content.decode()
        success = True
        try:
            timestamp = response.headers["BIZ_TIMESTAMP"]
            signature = response.headers["BIZ_RESP_SIGNATURE"]
            if self.debug:
                print(f"response <<<<<<<< \n content: {content}\n headers: {response.headers} \n")
            success = verify_ecdsa_signature("%s|%s" % (content, timestamp), signature, self.env.coboPub)
        except KeyError:
            pass
        return success, json.loads(content)

    def request(
            self,
            method: str,
            path: str,
            params: dict
    ) -> ApiResponse:
        method = method.upper()
        nonce = str(int(time.time() * 1000 * 1000))
        params = self.remove_none_value_elements(params)
        content = f"{method}|{path}|{nonce}|{self.sort_params(params)}"
        sign = self.api_signer.sign(content)

        headers = {
            "Biz-Api-Key": self.api_signer.get_public_key(),
            "Biz-Api-Nonce": nonce,
            "Biz-Api-Signature": sign,
        }
        url = f"{self.env.host}{path}"
        if self.debug:
            print(f"request >>>>>>>>\n method: {method} \n url: {url} \n params: {params} \n headers: {headers} \n")

        if method == "GET":
            resp = requests.get(url, params=urlencode(params), headers=headers)
        elif method == "POST":
            resp = requests.post(url, data=params, headers=headers)
        else:
            raise Exception("Not support http method")
        verify_success, result = self.verify_response(resp)
        if not verify_success:
            raise Exception("Fatal: verify content error, maybe encounter mid man attack")

        success = result['success']
        if success:
            return ApiResponse(True, result['result'], None)
        else:
            exception = ApiError(result['error_code'], result['error_message'], result['error_id'])
            return ApiResponse(False, None, exception)

    def get_supported_chains(self):
        params = {}
        return self.request("GET", "/v1/custody/mpc/get_supported_chains/", params)

    def get_supported_coins(self, chain_code: str) -> ApiResponse:
        params = {"chain_code": chain_code}
        return self.request("GET", "/v1/custody/mpc/get_supported_coins/", params)

    def get_supported_nft_collections(self, chain_code: str) -> ApiResponse:
        params = {"chain_code": chain_code}
        return self.request("GET", "/v1/custody/mpc/get_supported_nft_collections/", params)

    def get_wallet_supported_coins(self) -> ApiResponse:
        params = {}
        return self.request("GET", "/v1/custody/mpc/get_wallet_supported_coins/", params)

    def is_valid_address(self, coin: str, address: str) -> ApiResponse:
        params = {"coin": coin, "address": address}
        return self.request("GET", "/v1/custody/mpc/is_valid_address/", params)

    def get_main_address(self, chain_code: str) -> ApiResponse:
        params = {"chain_code": chain_code}
        return self.request("GET", "/v1/custody/mpc/get_main_address/", params)

    def generate_addresses(self, chain_code: str, count: int) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "count": count,
        }
        return self.request("POST", "/v1/custody/mpc/generate_addresses/", params)

    def list_addresses(self, chain_code: str, start_id: str = None, end_id: str = None, limit: int = None,
                       sort: int = None) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "start_id": start_id,
            "end_id": end_id,
            "limit": limit,
            "sort": sort
        }
        return self.request("GET", "/v1/custody/mpc/list_addresses/", params)

    def get_balance(self, address: str, chain_code: str = None, coin: str = None) -> ApiResponse:
        params = {
            "address": address,
            "chain_code": chain_code,
            "coin": coin
        }
        return self.request("GET", "/v1/custody/mpc/get_balance/", params)

    def list_balances(self, page_index: int, page_length: int, coin: str = None, chain_code: str = None) -> ApiResponse:
        params = {
            "coin": coin,
            "chain_code": chain_code,
            "page_index": page_index,
            "page_length": page_length
        }
        return self.request("GET", "/v1/custody/mpc/list_balances/", params)

    def list_spendable(self, coin: str, address: str = None) -> ApiResponse:
        params = {
            "address": address,
            "coin": coin,
        }
        return self.request("GET", "/v1/custody/mpc/list_spendable/", params)

    def create_transaction(self, coin: str, request_id: str, amount: str, from_addr: str = None,
                           to_addr: str = None, to_address_details: str = None, fee: float = None,
                           gas_price: int = None, gas_limit: int = None, operation: int = None,
                           extra_parameters: str = None, max_fee: int = None,
                           max_priority_fee: int = None, fee_amount: int = None) -> ApiResponse:
        params = {
            "coin": coin,
            "request_id": request_id,
            "from_address": from_addr,
            "to_address": to_addr,
            "amount": amount,
            "to_address_details": to_address_details,
            "fee": fee,
            "gas_price": gas_price,
            "gas_limit": gas_limit,
            "operation": operation,
            "extra_parameters": extra_parameters,
            "max_fee": max_fee,
            "max_priority_fee": max_priority_fee,
            "fee_amount": fee_amount,
        }
        return self.request("POST", "/v1/custody/mpc/create_transaction/", params)

    def sign_message(self, chain_code: str, request_id: str, from_addr: str,
                     sign_version: int, extra_parameters: str) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "request_id": request_id,
            "from_address": from_addr,
            "sign_version": sign_version,
            "extra_parameters": extra_parameters,
        }
        return self.request("POST", "/v1/custody/mpc/sign_message/", params)

    def drop_transaction(self, cobo_id: str, request_id: str, gas_price: int, gas_limit: int = None,
                         fee: float = None, fee_amount: int = None) -> ApiResponse:
        params = {
            "cobo_id": cobo_id,
            "gas_price": gas_price,
            "gas_limit": gas_limit,
            "request_id": request_id,
            "fee": fee,
            "fee_amount": fee_amount,
        }
        return self.request("POST", "/v1/custody/mpc/drop_transaction/", params)

    def speedup_transaction(self, cobo_id: str, request_id: str, gas_price: int = None, gas_limit: int = None,
                            fee: float = None, fee_amount: int = None) -> ApiResponse:
        params = {
            "cobo_id": cobo_id,
            "gas_price": gas_price,
            "gas_limit": gas_limit,
            "request_id": request_id,
            "fee": fee,
            "fee_amount": fee_amount,
        }
        return self.request("POST", "/v1/custody/mpc/speedup_transaction/", params)

    def transactions_by_request_ids(self, request_ids: str, status: int = None) -> ApiResponse:
        params = {"request_ids": request_ids, "status": status}
        return self.request("GET", "/v1/custody/mpc/transactions_by_request_ids/", params)

    def transactions_by_cobo_ids(self, cobo_ids: str, status: int = None) -> ApiResponse:
        params = {"cobo_ids": cobo_ids, "status": status}
        return self.request("GET", "/v1/custody/mpc/transactions_by_cobo_ids/", params)

    def transactions_by_tx_hash(self, tx_hash: str, transaction_type: int = None) -> ApiResponse:
        params = {"tx_hash": tx_hash, "transaction_type": transaction_type}
        return self.request("GET", "/v1/custody/mpc/transactions_by_tx_hash/", params)

    def list_transactions(self, start_time: int = None, end_time: int = None, status: int = None,
                          order: str = None, order_by: str = None, transaction_type: int = None,
                          coins: str = None,
                          from_address: str = None, to_address: str = None, limit: int = 50) -> ApiResponse:
        params = {
            "start_time": start_time,
            "end_time": end_time,
            "status": status,
            "order": order,
            "order_by": order_by,
            "transaction_type": transaction_type,
            "coins": coins,
            "from_address": from_address,
            "to_address": to_address,
            "limit": limit
        }
        return self.request("GET", "/v1/custody/mpc/list_transactions/", params)

    def estimate_fee(self, coin: str, amount: int = None, address: str = None, replace_cobo_id: str = None,
                     from_address: str = None,
                     to_address_details: str = None, fee: float = None, gas_price: int = None, gas_limit: int = None,
                     extra_parameters: str = None) -> ApiResponse:
        params = {"coin": coin, "amount": amount, "address": address, "replace_cobo_id": replace_cobo_id,
                  "from_address": from_address, "to_address_details": to_address_details, "fee": fee,
                  "gas_price": gas_price, "gas_limit": gas_limit, "extra_parameters": extra_parameters}
        return self.request("GET", "/v1/custody/mpc/estimate_fee/", params)

    def list_tss_node_requests(self, request_type: int = None, status: int = None) -> ApiResponse:
        params = {"request_type": request_type, "status": status}
        return self.request("GET", "/v1/custody/mpc/list_tss_node_requests/", params)

    def retry_double_check(self, request_id: str) -> ApiResponse:
        params = {"request_id": request_id}
        return self.request("POST", "/v1/custody/mpc/retry_double_check/", params)

    def list_tss_node(self) -> ApiResponse:
        return self.request("GET", "/v1/custody/mpc/list_tss_node/", {})

    def sign_messages_by_request_ids(self, request_ids: str) -> ApiResponse:
        params = {"request_ids": request_ids}
        return self.request("GET", "/v1/custody/mpc/sign_messages_by_request_ids/", params)

    def sign_messages_by_cobo_ids(self, cobo_ids: str) -> ApiResponse:
        params = {"cobo_ids": cobo_ids}
        return self.request("GET", "/v1/custody/mpc/sign_messages_by_cobo_ids/", params)



import json
import time
from typing import Tuple

from urllib.parse import urlencode

import requests

from cobo_custody.client.api_response import ApiResponse
from cobo_custody.config import Env
from cobo_custody.error.api_error import ApiError
from cobo_custody.signer.api_signer import ApiSigner
from cobo_custody.signer.local_signer import verify_ecdsa_signature


class Web3Client(object):
    def __init__(self, signer: ApiSigner, env: Env, debug: bool = False):
        self.api_signer = signer
        self.env = env
        self.debug = debug

    def sort_params(self, params: dict) -> str:
        params = [(key, val) for key, val in params.items()]

        params.sort(key=lambda x: x[0])
        return urlencode(params)

    def remove_none_value_elements(self, input_dict: dict) -> dict:
        if type(input_dict) is not dict:
            return {}
        result = {}
        for key in input_dict:
            tmp = {}
            if input_dict[key] is not None:
                if type(input_dict[key]).__name__ == 'dict':
                    tmp.update({key: self.remove_none_value_elements(input_dict[key])})
                else:
                    tmp.update({key: input_dict[key]})
            result.update(tmp)
        return result

    def verify_response(self, response: requests.Response) -> Tuple[bool, dict]:
        content = response.content.decode()
        success = True
        try:
            timestamp = response.headers["BIZ_TIMESTAMP"]
            signature = response.headers["BIZ_RESP_SIGNATURE"]
            if self.debug:
                print(f"response <<<<<<<< \n content: {content}\n headers: {response.headers} \n")
            success = verify_ecdsa_signature("%s|%s" % (content, timestamp), signature, self.env.coboPub)
        except KeyError:
            pass
        return success, json.loads(content)

    def request(
            self,
            method: str,
            path: str,
            params: dict
    ) -> ApiResponse:
        method = method.upper()
        nonce = str(int(time.time() * 1000 * 1000))
        params = self.remove_none_value_elements(params)
        content = f"{method}|{path}|{nonce}|{self.sort_params(params)}"
        sign = self.api_signer.sign(content)

        headers = {
            "Biz-Api-Key": self.api_signer.get_public_key(),
            "Biz-Api-Nonce": nonce,
            "Biz-Api-Signature": sign,
        }
        url = f"{self.env.host}{path}"
        if self.debug:
            print(f"request >>>>>>>>\n method: {method} \n url: {url} \n params: {params} \n headers: {headers} \n")

        if method == "GET":
            resp = requests.get(url, params=urlencode(params), headers=headers)
        elif method == "POST":
            resp = requests.post(url, data=params, headers=headers)
        else:
            raise Exception("Not support http method")
        verify_success, result = self.verify_response(resp)
        if not verify_success:
            raise Exception("Fatal: verify content error, maybe encounter mid man attack")

        success = result['success']
        if success:
            return ApiResponse(True, result['result'], None)
        else:
            exception = ApiError(result['error_code'], result['error_message'], result['error_id'])
            return ApiResponse(False, None, exception)

    def batch_web3_new_address(self, chain_code: str, count: int) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "count": count,
        }
        return self.request("POST", "/v1/custody/web3_add_addresses/", params)

    def get_web3_address_list(self, chain_code: str, page_index: int, page_length: int,
                              sort_flag: int = 0) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "page_index": page_index,
            "page_length": page_length,
            "sort_flag": sort_flag
        }
        return self.request("GET", "/v1/custody/web3_list_wallet_address/", params)

    def get_web3_wallet_asset_list(self, address: str = None, chain_code: str = None) -> ApiResponse:
        params = {
            "address": address,
            "chain_code": chain_code,
        }
        return self.request("GET", "/v1/custody/web3_list_wallet_assets/", params)

    def get_web3_wallet_nft_list(self, nft_code: str, address: str = None) -> ApiResponse:
        params = {
            "nft_code": nft_code,
            "address": address,
        }
        return self.request("GET", "/v1/custody/web3_list_wallet_nfts/", params)

    def get_web3_wallet_nft_detail(self, nft_code: str, token_id: str) -> ApiResponse:
        params = {
            "nft_code": nft_code,
            "token_id": token_id,
        }
        return self.request("GET", "/v1/custody/web3_wallet_nft_detail/", params)

    def get_web3_supported_chains(self) -> ApiResponse:
        params = {}
        return self.request("GET", "/v1/custody/web3_supported_chains/", params)

    def get_web3_supported_coins(self, chain_code: str) -> ApiResponse:
        params = {"chain_code": chain_code}
        return self.request("GET", "/v1/custody/web3_supported_coins/", params)

    def get_web3_supported_nft_collections(self) -> ApiResponse:
        params = {}
        return self.request("GET", "/v1/custody/web3_supported_nft_collections/", params)

    def get_web3_supported_contracts(self, chain_code: str) -> ApiResponse:
        params = {"chain_code": chain_code}
        return self.request("GET", "/v1/custody/web3_supported_contracts/", params)

    def get_web3_supported_contract_methods(self, chain_code: str, contract_address: str) -> ApiResponse:
        params = {"chain_code": chain_code, "contract_address": contract_address}
        return self.request("GET", "/v1/custody/web3_supported_contract_methods/", params)

    def web3_withdraw(self, coin: str, request_id: str, from_addr: str, to_addr: str, amount: int) -> ApiResponse:
        params = {"coin": coin, "request_id": request_id, "from_addr": from_addr, "to_addr": to_addr, "amount": amount}
        return self.request("POST", "/v1/custody/web3_withdraw/", params)

    def get_web3_withdraw_transaction(self, request_id: str) -> ApiResponse:
        params = {"request_id": request_id}
        return self.request("GET", "/v1/custody/web3_get_withdraw_transaction/", params)

    def web3_contract(self, chain_code: str, request_id: str, wallet_addr: str, contract_addr: str, method_id: str,
                      method_name: str, args: str, amount: int = 0, gas_limit: int = None) -> ApiResponse:
        params = {
            "chain_code": chain_code,
            "request_id": request_id,
            "wallet_addr": wallet_addr,
            "contract_addr": contract_addr,
            "method_id": method_id,
            "method_name": method_name,
            "args": args,
            "amount": amount,
            "gas_limit": gas_limit,
        }
        return self.request("POST", "/v1/custody/web3_contract/", params)

    def get_web3_contract_transaction(self, request_id: str) -> ApiResponse:
        params = {"request_id": request_id}
        return self.request("GET", "/v1/custody/web3_get_contract_transaction/", params)

    def list_web3_wallet_transactions(self, address: str, chain_code: str = None, max_id: str = None,
                                      min_id: str = None,
                                      limit: int = 50) -> ApiResponse:
        params = {"address": address, "chain_code": chain_code, "max_id": max_id, "min_id": min_id, "limit": limit}
        return self.request("GET", "/v1/custody/web3_list_wallet_transactions/", params)
